-module(eRum_app).

-behaviour(application).

-include("eRum.hrl").

-export([
   start/0
   , start/2
   , stop/1
   , doStart/1
   , startHandler/3
   , configure_sink/2
]).



start() ->
   application:ensure_all_started(eRum).

start(_StartType, _StartArgs) ->
   {ok, Pid} = eRum_sup:start_link(),
   SavedHandlers = boot(),
   _ = doStart('__all_extra'),
   _ = doStart('__traces'),
   clean_up_config_checks(),
   {ok, Pid, SavedHandlers}.

boot() ->
   %% Handle the default sink.
   tryStartAsyncMgr(rumUtil:get_env(asyncThreshold, undefined), rumUtil:get_env(asyncThresholdWindow, undefined), ?RumDefSink),

   tryInstallKiller(rumUtil:get_env(killerHwm, undefined), rumUtil:get_env(killerReTime, undefined), ?RumDefSink),

   tryStartHandlers(rumUtil:get_env(handlers, ?RumDefHandler), ?RumDefSink),



   SavedHandlers = start_error_logger_handler(rumUtil:get_env(error_logger_redirect, true), rumUtil:get_env(error_logger_hwm, 0), rumUtil:get_env(error_logger_whitelist, [])),
   eRum:update_loglevel_config(?RumDefSink),
   SavedHandlers.

doStart('__traces') ->
   _ = rumUtil:trace_filter(none),
   ok = add_configured_traces();

doStart('__all_extra') ->
   configure_extra_sinks(application:get_env(lager, extra_sinks, []));

doStart(?RumDefSink) -> boot();
doStart(Sink) ->
   AllSinksDef = application:get_env(lager, extra_sinks, []),
   boot_sink(Sink, lists:keyfind(Sink, 1, AllSinksDef)).

boot_sink(Sink, {Sink, Def}) ->
   configure_sink(Sink, Def);
boot_sink(Sink, false) ->
   configure_sink(Sink, []).

stop(Handlers) ->
   lists:foreach(fun(Handler) ->
      error_logger:add_report_handler(Handler)
                 end, Handlers),
   rumConfig:cleanup().

tryStartAsyncMgr(undefined, _Window, _SinkS) ->
   ignore;
tryStartAsyncMgr(Threshold, Window, Sink) ->
   case Window of
      undefined ->
         _ = supervisor:start_child(rumHWatcherSup, [Sink, ?RumBackendThrottle, [Threshold, erlang:trunc(Threshold * 0.2)]]);
      _ ->
         _ = supervisor:start_child(rumHWatcherSup, [Sink, ?RumBackendThrottle, [Threshold, Window]])
   end,
   ok.

tryInstallKiller(undefined, _ReTime, _Sink) -> ok;
tryInstallKiller(HWM, ReTime, Sink) ->
   case ReTime of
      undefined ->
         _ = supervisor:start_child(rumHWatcherSup, [Sink, rumMgrKiller, [HWM, 5000]]);
      _ ->
         _ = supervisor:start_child(rumHWatcherSup, [Sink, rumMgrKiller, [HWM, ReTime]])
   end,
   ok.

tryStartHandlers(undefined, _Sink) -> ok;
tryStartHandlers(Handlers, Sink) ->
   %% 启动失败的处理程序将在handler_watcher中处理
   NewHandler = doStartHandlers(Handlers, Sink, [], []),
   rumConfig:global_set(handlers, rumConfig:global_get(handlers, []) ++ NewHandler),
   ok.

doStartHandlers([], _Sink, _NameAcc, HandlerAcc) ->
   HandlerAcc;
doStartHandlers([OneHandler | Handlers], Sink, NameAcc, HandlerAcc) ->
   {Module, Options} = expand_handlers(OneHandler),
   NewNameAcc =
      case Module of
         {rumBackendFile, F} ->
            case lists:member(F, NameAcc) of
               true ->
                  error_logger:error_msg("Cannot have same file (~p) in multiple file backends~n", [F]),
                  throw({error, bad_config});
               _ ->
                  [F | NameAcc]
            end;
         _ ->
            NameAcc
      end,

   HandlerRet = startHandler(Sink, Module, Options),
   doStartHandlers(Handlers, Sink, NewNameAcc, [HandlerRet | HandlerAcc]).

startHandler(Sink, Module, Config) ->
   {ok, Watcher} = supervisor:start_child(rumHWatcherSup, [Sink, Module, Config]),
   {Module, Watcher, Sink}.

interpret_hwm(undefined) ->
   undefined;
interpret_hwm(HWM) when not is_integer(HWM) orelse HWM < 0 ->
   _ = eRum:log(warning, self(), "Invalid error_logger high water mark: ~p, disabling", [HWM]),
   undefined;
interpret_hwm(HWM) ->
   HWM.

-spec start_error_logger_handler(boolean(), pos_integer(), list()) -> list().
start_error_logger_handler(false, _HWM, _Whitelist) ->
   [];
start_error_logger_handler(true, HWM, WhiteList) ->
   GlStrategy =
      case application:get_env(lager, error_logger_groupleader_strategy) of
         undefined ->
            handle;
         {ok, GlStrategy0} when
            GlStrategy0 =:= handle;
            GlStrategy0 =:= ignore;
            GlStrategy0 =:= mirror ->
            GlStrategy0;
         {ok, BadGlStrategy} ->
            error_logger:error_msg(
               "Invalid value for 'error_logger_groupleader_strategy': ~p~n",
               [BadGlStrategy]),
            throw({error, bad_config})
      end,

   case whereis(error_logger) of
      undefined ->
         %% On OTP 21 and above, error_logger is deprecated in favor of 'logger'
         %% As a band-aid, boot up error_logger anyway and install it as a logger handler
         %% we can't use error_logger:add_report_handler because we want supervision of the handler
         %% so we have to manually add the logger handler
         %%
         %% Longer term we should be installing a logger handler instead, but this will bridge the gap
         %% for now.
         _ = error_logger:start(),
         _ = logger:add_handler(error_logger, error_logger, #{level => info, filter_default => log}),
         ok = maybe_remove_logger_handler();
      _ ->
         ok
   end,

   %% capture which handlers we removed from error_logger so we can restore them when lager stops
   OldHandlers =
      case supervisor:start_child(lager_handler_watcher_sup, [error_logger, error_logger_lager_h, [HWM, GlStrategy]]) of
         {ok, _} ->
            [begin error_logger:delete_report_handler(X), X end ||
               X <- gen_event:which_handlers(error_logger) -- [error_logger_lager_h | WhiteList]];
         {error, _} ->
            []
      end,
   OldHandlers.

%% On OTP 21.1 and higher we need to remove the `default' handler.
%% But it might not exist, so we will wrap this in a try-catch
%% block
maybe_remove_logger_handler() ->
   try
      ok = logger:remove_handler(default)
   catch
      error:undef -> ok;
      Err:Reason ->
         error_logger:error_msg("calling logger:remove_handler(default) failed: ~p ~p",
            [Err, Reason])
   end.

configure_sink(Sink, SinkDef) ->
   rumConfig:initSink(Sink),
   ChildId = rumUtil:make_internal_sink_name(Sink),
   _ = supervisor:start_child(eRum_sup, {ChildId, {gen_event, start_link, [{local, Sink}]}, permanent, 5000, worker, dynamic}),
   tryStartAsyncMgr(Sink, proplists:get_value(asyncThreshold, SinkDef), proplists:get_value(asyncThresholdWindow, SinkDef)),
   _ = tryInstallKiller(Sink, proplists:get_value(killerHwm, SinkDef),
      proplists:get_value(killerReTime, SinkDef)),
   tryStartHandlers(proplists:get_value(handlers, SinkDef, []), Sink),
   eRum:update_loglevel_config(Sink).


configure_extra_sinks(Sinks) ->
   lists:foreach(fun({Sink, Proplist}) -> configure_sink(Sink, Proplist) end, Sinks).



expand_handlers([]) ->
   [];
expand_handlers([{lager_file_backend, [{Key, _Value} | _] = Config} | T]) when is_atom(Key) ->
   %% this is definitely a new-style config, no expansion needed
   [maybe_make_handler_id(lager_file_backend, Config) | expand_handlers(T)];
expand_handlers([{lager_file_backend, Configs} | T]) ->
   ?INT_LOG(notice, "Deprecated lager_file_backend config detected, please consider updating it", []),
   [{rumBackendFile:config_to_id(Config), Config} || Config <- Configs] ++
   expand_handlers(T);
expand_handlers([{Mod, Config} | T]) when is_atom(Mod) ->
   [maybe_make_handler_id(Mod, Config) | expand_handlers(T)];
expand_handlers([H | T]) ->
   [H | expand_handlers(T)].

add_configured_traces() ->
   Traces = case application:get_env(lager, traces) of
               undefined ->
                  [];
               {ok, TraceVal} ->
                  TraceVal
            end,

   lists:foreach(fun start_configured_trace/1, Traces),
   ok.

start_configured_trace({Handler, Filter}) ->
   {ok, _} = eRum:trace(Handler, Filter);
start_configured_trace({Handler, Filter, Level}) when is_atom(Level) ->
   {ok, _} = eRum:trace(Handler, Filter, Level).

maybe_make_handler_id(Mod, Config) ->
   %% Allow the backend to generate a gen_event handler id, if it wants to.
   %% We don't use erlang:function_exported here because that requires the module
   %% already be loaded, which is unlikely at this phase of startup. Using code:load
   %% caused undesirable side-effects with generating code-coverage reports.
   try Mod:config_to_id(Config) of
      Id ->
         {Id, Config}
   catch
      error:undef ->
         {Mod, Config}
   end.